"""
The action module is responsible for containing one Python module for each action which Cobbler offers. The code should
never be dependent on another module or on other parts. An action should request the exact
data it requires and nothing more.
"""
