"""
This package contains all data storage classes. The classes are responsible for ensuring that types of the properties
are correct but not for logical checks. The classes should be as stupid as possible. Further they are responsible for
returning the logic for serializing and deserializing themselves.
"""
